@@ -7,8 +7,8 @@ import shutil
 import socket
 import getpass
 import pkgutil
+import logging
 import platform
-import tempfile
 import functools
 import subprocess
 import virtualenv
@@ -36,13 +36,14 @@ try:
 except ImportError:
     from importlib_metadata import Distribution
 
-from ._version import __version__
+from . import __version__
 from .launch import shell
 from .recipe import ContainerRecipe, RevisionRecipe, DEFAULT_CONTAINER_NAME
 from .exceptions import ContainerError
 
 
 _PY2 = sys.version_info.major == 2
+_log = logging.getLogger("rezup")
 
 
 # TODO:
@@ -52,6 +53,12 @@ _PY2 = sys.version_info.major == 2
 #   - [ ] REZUP_CLEAN_AFTER
 
 
+if _PY2:
+    string_types = basestring,  # noqa
+else:
+    string_types = str,
+
+
 def makedirs(path):
     path = str(path)
     if not os.path.isdir(path):
@@ -68,10 +75,11 @@ def norm_path(path):
 
 
 def iter_containers():
-    """Iterate containers by recipes (~/rezup[.{name}].toml)
+    """Iterate containers by recipes (`~/rezup[.{name}].toml`)
 
-    Yields:
-        Container
+    Returns:
+        Generator[rezup.Container]: Instance of [rezup.Container][], could be
+            linked to remote or local.
 
     """
     for recipe in ContainerRecipe.iter_recipes():
@@ -84,6 +92,29 @@ def iter_containers():
 
 
 def get_container_root(recipe, remote=False):
+    """Get container root path from recipe
+
+    ```toml
+    # rezup.foo.toml
+    [root]
+    local = false
+    remote = "/path/to/remote/containers"
+
+    ```
+
+    If the `recipe` represents the above example recipe file, container named
+    `foo` will have local root set to `rezup.recipe.DEFAULT_CONTAINER_RECIPES`,
+    unless environ variable `REZUP_ROOT_LOCAL` has value. And the remote root
+    will set to `/path/to/remote/containers`.
+
+    Args:
+        recipe (`ContainerRecipe`): The container recipe to lookup from.
+        remote (bool): Fetch remote root if True, otherwise local.
+
+    Returns:
+        `pathlib.Path`: Container root path, local or remote one.
+
+    """
     data = recipe.data()
     if remote:
         remote = \
@@ -108,7 +139,7 @@ class Container:
     affecting any existing consumer.
 
     In filesystem, a container is a folder that has at least one Rez venv
-    installation exists, and those Rez venv folders (revisions) are named
+    installation exists, and those Rez venv folders (`Revision`) are named
     by timestamp, so when the container is being asked for a venv to use,
     the current latest available one will be sorted out.
 
@@ -120,8 +151,17 @@ class Container:
     contains the venv installation manifest file, and when being asked, a
     venv will be created locally or re-used if same revision exists in local.
 
+    Args:
+        name (`str`, optional): Container name, use `Container.DEFAULT_NAME`
+            if not given.
+        recipe (`ContainerRecipe`, optional): A `ContainerRecipe` object to
+            help constructing container, will look into `ContainerRecipe.RECIPES_DIR`
+            if not given.
+        force_local (`bool`, optional): Default `False`. Ignore linking remote
+            even if the remote path has set, always link to local if `True`.
+
     """
-    DEFAULT_NAME = DEFAULT_CONTAINER_NAME
+    DEFAULT_NAME = DEFAULT_CONTAINER_NAME  #: `rezup.recipe.DEFAULT_CONTAINER_NAME`
 
     def __init__(self, name=None, recipe=None, force_local=False):
         name = name or self.DEFAULT_NAME
@@ -140,44 +180,93 @@ class Container:
         self._root = root
         self._path = root / name
 
+    def __repr__(self):
+        return "%s(remote=%d, name=%r, path=%r)" % (
+            self.__class__.__name__,
+            int(self.is_remote()),
+            self.name(),
+            self._path.resolve(),
+        )
+
     @classmethod
     def create(cls, name, force_local=False):
+        """Create container from recipe.
+
+        Args:
+            name (str): The name of container.
+            force_local (bool): Create as local if True, or try remote.
+
+        Returns:
+            `Container`: A container instance.
+        """
         recipe = ContainerRecipe(name)
+        _log.debug("Sourcing recipe from: %s" % recipe)
         if not recipe.is_file():
+            _log.debug("Recipe file not exists, creating default..")
             recipe.create()
         return Container(name, recipe, force_local)
 
     def root(self):
-        """Root path of current container"""
+        """
+        Returns:
+            `pathlib.Path`: Path to where this container is located.
+        """
         return self._root
 
     def name(self):
-        """Name of current container"""
+        """
+        Returns:
+            str: Container name.
+        """
         return self._name
 
     def path(self):
-        """Path of current container"""
+        """
+        Returns:
+            `pathlib.Path`: Full path of this container.
+        """
         return self._path
 
     def recipe(self):
-        """Returns an Recipe instance that binds to this container"""
+        """
+        Returns:
+            `ContainerRecipe`: The recipe instance that binds to this container.
+        """
         return self._recipe
 
     def libs(self):
-        """Root path of current container's shared libraries"""
+        """
+        Returns:
+            `pathlib.Path`: Path to this container's shared libraries.
+        """
         return self._path / "libs"
 
     def revisions(self):
-        """Root path of current container's revisions"""
+        """
+        Returns:
+            `pathlib.Path`: Path to this container's revisions.
+        """
         return self._path / "revisions"
 
     def is_exists(self):
+        """
+        Returns:
+            bool: Return `True` if this container exists.
+        """
         return self._path.is_dir()
 
     def is_empty(self):
+        """
+        Returns:
+            bool: Return `True` if this container has no valid revisions.
+        """
         return not bool(next(self.iter_revision(), None))
 
     def is_remote(self):
+        """
+        Returns:
+            bool: Return `True` if this container is linked to remote side.
+        """
         return self._remote
 
     def purge(self):
@@ -198,36 +287,92 @@ class Container:
             rmtree(root)
 
     def iter_revision(self, validate=True, latest_first=True):
+        """Iterating revisions from this container.
+
+        Args:
+            validate (bool, optional): Default `True`. Only yield revisions
+                that are valid if `True`.
+            latest_first (bool): Default `True`. Yield revisions based on
+                it's directory name (timestamp string) in descending order,
+                or ascending if `False`.
+
+        Yields:
+            Revision: `Revision` instances that match the condition.
+
+        """
+        _log.debug("Iterating revisions in container %s.." % self)
+
         if not self.is_exists():
+            _log.debug("Container %r not exists." % self.name())
             return
 
         revisions_root = Revision.compose_path(container=self)
         if not revisions_root.is_dir():
+            _log.debug("Revision root is not a directory: %s" % revisions_root)
             return
 
         revisions_root = str(revisions_root)
         for entry in sorted(os.listdir(revisions_root), reverse=latest_first):
             revision = Revision(container=self, dirname=entry)
+            _log.debug("... %s" % revision)
             if not validate or revision.is_valid():
                 yield revision
 
     def get_latest_revision(self, only_ready=True):
+        """Get latest revision from this container.
+
+        Args:
+            only_ready (bool): Default `True`. Include revisions that are not
+                in ready state if `False`.
+
+        Returns:
+            Revision: An instance of `Revision` if found, or `None`.
+
+        """
         for revision in self.iter_revision():
             if not only_ready or revision.is_ready():
+                _log.debug("Found latest revision.")
                 return revision
 
-    def get_revision_at_time(self, timestamp, strict=False, only_ready=True):
+    def get_revision_by_time(self, timestamp, fallback=False, only_ready=True):
+        """Returns a revision that match the timestamp
+
+        Args:
+            timestamp (datetime.datetime): a time for matching revision
+            fallback (bool): If True, accept earlier revision when no exact matched
+            only_ready (bool): Include revisions that are not in ready state if False
+
+        Returns:
+            Revision: An instance of `Revision` if found, or `None`.
+
+        """
         for revision in self.iter_revision():
             if not only_ready or revision.is_ready():
                 if (revision.timestamp() == timestamp
-                        or (not strict and revision.timestamp() <= timestamp)):
+                        or (fallback and revision.timestamp() <= timestamp)):
+
+                    _log.debug("Found time matched revision.")
                     return revision
 
+        _log.debug("No time matched revision found.")
+
     def new_revision(self):
+        """Create a new revision
+
+        Returns:
+            Revision: An instance of `Revision` that just created.
+
+        """
         return Revision.create(self)
 
 
 class Revision:
+    """
+    Args:
+        container (Container): The container that holds this revision
+        dirname (str, optional): Directory name of this revision, should be
+            a timestamp string if given.
+    """
 
     def __init__(self, container, dirname=None):
         dirname = str(dirname or time.time())
@@ -240,6 +385,23 @@ class Revision:
         self._metadata = None
         self._recipe = RevisionRecipe(self)
         self._metadata_path = self._path / "revision.json"
+        self._is_pulled = False
+
+    def __repr__(self):
+        return "%s(valid=%d, ready=%d, remote=%d, time=%s, path=%r)" % (
+            self.__class__.__name__,
+            int(self.is_valid()),
+            int(self.is_ready()),
+            int(self.is_remote()),
+            self.time_str() or "?",
+            self._path.resolve(),
+        )
+
+    def __eq__(self, other):
+        if not isinstance(other, type(self)):
+            return False
+        return self._container.name() == other._container.name() \
+            and self.timestamp() == other.timestamp()
 
     @classmethod
     def compose_path(cls, container, dirname=None):
@@ -255,14 +417,23 @@ class Revision:
         return revision
 
     def _write(self, pulling=None):
+        """Sourcing recipe and create a revision
+
+        Args:
+            pulling (Revision, optional): If given, pulling recipe from that
+                revision, usually a remote one.
+        """
+        _log.info("Creating revision..")
+
         recipe = self._recipe
         makedirs(self._path)
 
         # write revision recipe
         if pulling is None:
-            print("Recipe sourced from: %s" % self._container.recipe().path())
+            _log.info("Recipe sourced from: %s" % self._container.recipe())
             recipe.create()
         else:
+            _log.debug("Pulling recipe from: %s" % pulling)
             recipe.pull(pulling)
 
         if not self.is_valid():
@@ -272,6 +443,10 @@ class Revision:
         rez_ = Tool(recipe["rez"])
         extensions = [Tool(d) for d in recipe.get("extension", []) if d]
         shared_lib = recipe.get("shared")
+        _log.info("Recipe loaded.")
+        _log.debug("       Rez: %s" % rez_.url)
+        _log.debug(" Extension: " + ", ".join([e.name for e in extensions]))
+        _log.debug("Shared-lib: %s" % shared_lib)
 
         # install, if at local
         if not self._container.is_remote():
@@ -288,9 +463,13 @@ class Revision:
                 "venvs": ["rez"] + [t.name for t in extensions if t.isolation],
             }, indent=4))
 
+        _log.info("Revision created: %s" % self)
+
     def _install(self, rez_, extensions=None, shared_lib=None):
         """Construct Rez virtual environment by recipe
         """
+        _log.debug("Installing..")
+
         extensions = extensions or []
         installer = Installer(self)
         installer.install_rez(rez_)
@@ -335,6 +514,10 @@ class Revision:
     def timestamp(self):
         return self._timestamp
 
+    def time_str(self):
+        return self._timestamp.strftime("%d.%b.%y %H:%M") \
+            if self._timestamp else ""
+
     def container(self):
         return self._container
 
@@ -362,7 +545,7 @@ class Revision:
 
         def file_loader(d):
             return [d[k] for k in sorted([
-                k for k, v in d.items() if isinstance(v, str)])]
+                k for k, v in d.items() if isinstance(v, string_types)])]
 
         dot_env = recipe.get("dotenv")
         if dot_env:
@@ -386,6 +569,7 @@ class Revision:
 
         env.update({
             "REZUP_CONTAINER": self._container.name(),
+            "REZUP_USING_REMOTE": "yes" if self._is_pulled else "",
         })
 
         return env
@@ -406,7 +590,7 @@ class Revision:
             if revision.timestamp() > self._timestamp:
                 yield revision
 
-    def pull(self, check_out=True):
+    def pull(self, check_out=True, fallback=False):
         """Return corresponding local side revision
 
         If the revision is from remote container, calling this method will
@@ -416,9 +600,10 @@ class Revision:
         If the revision is from local container, return `self`.
 
         Args:
-            check_out(bool, optional): When no matched local revision,
+            check_out (bool, optional): When no matched local revision,
                 create one if True or just return None at the end.
                 Default is True.
+            fallback (bool, optional)
 
         Returns:
             Revision or None
@@ -430,68 +615,116 @@ class Revision:
         # get local
         _con_name = self._container.name()
         _con_recipe = self._container.recipe()  # careful, this affect's root
+
         local = Container(_con_name, recipe=_con_recipe, force_local=True)
-        revision = local.get_revision_at_time(self._timestamp, strict=True)
-        if revision is None and check_out:
-            print("Pulling from remote container %s .." % _con_name)
-            revision = Revision(container=local, dirname=self._dirname)
-            revision._write(pulling=self)
+        rev = local.get_revision_by_time(self._timestamp,
+                                         fallback=fallback,
+                                         only_ready=True)
 
-        return revision
+        _allow_create = rev is None and check_out
+        _did_fallback = rev.timestamp() != self._timestamp if rev else False
 
-    def use(self, run_script=None):
+        if not fallback and _allow_create:
+            _log.info("Pulling from remote container: %s" % self._container)
+            rev = Revision(container=local, dirname=self._dirname)
+            rev._write(pulling=self)
+
+        if fallback and _did_fallback:
+            _log.warning(
+                "Local revision at exact time (%s) is not ready, "
+                "fallback to %s" % (self.time_str(), rev)
+            )
+
+        if rev is not None:
+            rev._is_pulled = True
+
+        return rev
+
+    def spawn_shell(self, command=None):
+        """Spawn a sub-shell
+
+        Args:
+            command (list, optional): Shell script file with args or commands.
+                If given, the sub-shell will not be interactive.
+
+        Returns:
+            subprocess.Popen
+
+        Raises:
+            ContainerError
+
+        """
         if not self.is_valid():
-            raise Exception("Cannot use invalid revision.")
+            raise ContainerError("Cannot use invalid revision.")
         if not self.is_ready():
-            raise Exception("Revision is not ready to be used.")
+            raise ContainerError("Revision is not ready to be used.")
 
         if self.is_remote():
             # use local
             revision = self.pull()
-            return revision.use(run_script=run_script)
+            if revision is None:
+                raise ContainerError("No revision pulled.")
+            if not revision.is_ready():
+                raise ContainerError("Revision is not ready to be used.")
+
+            return revision.spawn_shell(command=command)
 
         else:
             # Launch subprocess
-            replacement = dict()
             environment = self._compose_env()
             shell_name, shell_exec = self._get_shell()
 
-            if run_script and os.path.isfile(run_script):
-                # run shell script and exit
-                block = False
-                replacement["__REZUP_DO__"] = "script"
-                replacement["__REZUP_DO_SCRIPT__"] = os.path.abspath(run_script)
-
-            elif run_script:
-                # run shell command and exit
-                block = False
-                replacement["__REZUP_DO__"] = "command"
-                replacement["__REZUP_DO_COMMAND__"] = run_script
+            if command:
+                # run command and exit
+                if command[0] == ".":
+                    cmd = command
+                else:
+                    exe = command[0]
+                    exe = shell.which(exe, env=environment) or exe
+                    cmd = [exe] + command[1:]
 
             else:
                 # interactive shell
-                block = True
-                replacement["__REZUP_SHELL__"] = shell_exec
-                prompt = "rezup (%s) " % self._container.name()
+                _con_name = self._container.name()
+                _con_from = "remote" if self._is_pulled else "local"
+                prompt = "rezup (%s/%s) " % (_con_name, _con_from)
                 prompt = shell.format_prompt_code(prompt, shell_name)
                 environment.update({
                     "REZUP_PROMPT": os.getenv("REZUP_PROMPT", prompt),
                 })
 
-            launch_script = shell.generate_launch_script(
-                shell_name,
-                dst_dir=tempfile.mkdtemp(),
-                replacement=replacement
-            )
-            cmd = shell.get_launch_cmd(shell_name,
-                                       shell_exec,
-                                       launch_script,
-                                       block=block)
+                cmd = shell.get_launch_cmd(
+                    shell_name,
+                    shell_exec,
+                    interactive=True,
+                )
 
             popen = subprocess.Popen(cmd, env=environment)
-            stdout, stderr = popen.communicate()
 
+            return popen
+
+    def use(self, command=None, wait=True):
+        """Run a sub-shell
+
+        Args:
+            command (list, optional): Shell script with args or commands. If
+                given, the sub-shell will not be interactive.
+            wait (bool, optional): Whether to wait `command` finish or not,
+                default True.
+
+        Returns:
+            int: subprocess return code, will always return 0 if `command`
+                is given and `wait` is False.
+
+        """
+        block = not command
+        popen = self.spawn_shell(command=command)
+
+        if block or wait:
+            stdout, stderr = popen.communicate()
             return popen.returncode
+        else:
+            return 0
 
     def _compose_env(self):
         env = os.environ.copy()
@@ -538,7 +771,12 @@ class Revision:
 
     @_require_local  # noqa
     def locate_rez_lib(self, venv_session=None):
-        """Returns rez module location in this revision"""
+        """Returns rez module location in this revision
+
+        Returns:
+            pathlib.Path or None if not found.
+
+        """
         if venv_session is None:
             venv_path = self.path() / "venv" / "rez"
             venv_session = virtualenv.session_via_cli(args=[str(venv_path)])
@@ -564,8 +802,16 @@ class Revision:
 
     @_require_local  # noqa
     def get_rez_version(self, venv_session=None):
-        """Returns rez version installed in this revision"""
+        """Returns rez version installed in this revision
+
+        Returns:
+            str or None if not found.
+
+        """
         rez_location = self.locate_rez_lib(venv_session)
+        if rez_location is None:
+            return
+
         version_py = rez_location / "rez" / "utils" / "_version.py"
         if version_py.is_file():
             _locals = {"_rez_version": ""}
@@ -575,7 +821,12 @@ class Revision:
 
     @_require_local  # noqa
     def production_bin_dir(self, venv_name):
-        """Returns production bin scripts dir in this revision"""
+        """Returns production bin scripts dir in this revision
+
+        Returns:
+            pathlib.Path, but the directory may not exists.
+
+        """
         bin_dirname = "Scripts" if platform.system() == "Windows" else "bin"
         venv_bin_dir = self.path() / "venv" / venv_name / bin_dirname
         return venv_bin_dir / "rez"
@@ -587,9 +838,11 @@ class Revision:
         metadata = self.metadata()
         if metadata and not metadata.get("rezup_version"):
             # rezup-1.x
+            _log.debug("Is a 'rezup-1.x' styled revision.")
             bin_dirs.append(self._path / "bin")
         else:
             # rezup-2.x
+            _log.debug("Is a 'rezup-2.x' styled revision.")
             for venv_name in metadata.get("venvs", []):
                 bin_dirs.append(self.production_bin_dir(venv_name))
 
@@ -604,6 +857,17 @@ class Tool:
         self.edit = data.get("edit", False)
         self.isolation = data.get("isolation", False)
         self.python = data.get("python", None)
+        self.flags = data.get("flags", ["-E"])
+
+    def __repr__(self):
+        return "%s(name=%s, edit=%d, url=%s, isolation=%d, python=%s)" % (
+            self.__class__.__name__,
+            self.name,
+            int(self.edit),
+            self.url,
+            int(self.isolation),
+            self.python,
+        )
 
 
 class Installer:
@@ -644,6 +908,10 @@ class Installer:
         use_python = tool.python or sys.executable
         dst = self._revision.path() / "venv" / tool.name
 
+        _log.info("Creating virtual env..")
+        _log.info("  Python: %s" % use_python)
+        _log.info("       -> %s" % dst)
+
         # create venv
         args = [
             "--python", str(use_python),
@@ -667,7 +935,7 @@ class Installer:
             "--disable-pip-version-check",
         ]
 
-        print("Installing %s.." % tool.name)
+        _log.info("Installing %s.." % tool)
         subprocess.check_output(cmd)
 
         if patch_scripts:
@@ -677,6 +945,8 @@ class Installer:
                 # TODO: copy completion scripts
 
     def mark_as_rez_production_install(self, tool, venv_session):
+        _log.info("Mark as Rez production install..")
+
         rez_bin = self._revision.production_bin_dir("rez")
         validator = rez_bin / ".rez_production_install"
         rez_version = self._revision.get_rez_version(venv_session)
@@ -700,6 +970,8 @@ class Installer:
         which will ignore all PYTHON* env vars, e.g. PYTHONPATH and PYTHONHOME.
 
         """
+        _log.info("Generating production scripts..")
+
         site_packages = venv_session.creator.purelib
         bin_path = venv_session.creator.bin_dir
 
@@ -749,12 +1021,14 @@ class Installer:
 
         scripts = maker.make_multiple(
             specifications=specifications.values(),
-            options=dict(interpreter_args=["-E"])
+            options=dict(interpreter_args=list(tool.flags))
         )
 
         return scripts
 
     def create_shared_lib(self, name, requires):
+        _log.info("Installing shared libs..")
+
         lib_path = str(self._container.libs() / name)
 
         venv_session = self._default_venv
